# Copyright 2015-2020 Canonical Ltd.
#
# This file is part of the Telegraf Charm for Juju.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3, as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranties of
# MERCHANTABILITY, SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR
# PURPOSE.  See the GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import itertools as it
import logging
import re
import time
import unittest

import requests

from zaza import model
from zaza.utilities import juju

DEFAULT_HTTPGET_TIMEOUT = 10
DEFAULT_RETRIES = 12
DEFAULT_RETRIES_TIMEOUT = 5
DEFAULT_TELEGRAF_EXPORTER_PORT = "9103"
DEB_TELEGRAF_CONF = "/etc/telegraf/telegraf.conf"
SNAP_TELEGRAF_CONF = "/var/snap/telegraf/current/telegraf.conf"


class BaseTelegrafTest(unittest.TestCase):
    """Base for Telegraf charm tests."""

    @classmethod
    def setUpClass(cls):
        super(BaseTelegrafTest, cls).setUpClass()
        cls.model_name = model.get_juju_model()
        cls.application_name = "telegraf"
        cls.lead_unit_name = model.get_lead_unit_name(
            cls.application_name, model_name=cls.model_name
        )
        logging.debug("Leader unit is {}".format(cls.lead_unit_name))


class TestTelegraf(BaseTelegrafTest):
    def check_re_pattern(self, re_pattern, text):
        logging.info("checking metrics %s", re_pattern)
        # findall returns a list, [] when no match
        self.assertTrue(re.findall(re_pattern, text, flags=re.M))

    def test_01_service(self):
        principal_units = (
            model.get_units(app)
            for app in juju.get_principle_applications(self.application_name)
        )

        for unit in it.chain.from_iterable(principal_units):
            if not unit.public_address:
                continue
            url = "http://{}:{}/metrics".format(
                unit.public_address, DEFAULT_TELEGRAF_EXPORTER_PORT
            )

            for retry in range(DEFAULT_RETRIES):
                resp = requests.get(url, timeout=DEFAULT_HTTPGET_TIMEOUT)
                self.assertEqual(resp.status_code, 200)

                if (
                    unit.name.startswith("postgresql/")
                    and "postgresql_blks_hit" not in resp.text  # noqa W503
                ) or "cpu_usage_idle" not in resp.text:
                    time.sleep(DEFAULT_RETRIES_TIMEOUT)

                    continue

                logging.info("test_service: Unit {} is reachable".format(unit.name))

                resp = requests.get(url)
                self.assertTrue(resp.ok)
                text = resp.text

                # check metrics are available
                re_patterns = [
                    r"^sockstat_",
                    r"^sockstat6_",
                    r"^softnet_stat",
                    r"^buddyinfo_",
                    r"^zoneinfo_",
                    r"^processes_",
                ]

                for re_pattern in re_patterns:
                    self.check_re_pattern(re_pattern, text)

                break
            else:
                msg = "test_service: Unit {} is NOT reachable".format(unit.name)
                logging.error(msg)
                self.fail(msg)

    def test_02_telegraf_logfile(self):
        """Check logfile parameter exists in ${BASE_DIR}/telegraf.conf."""
        install_method = model.get_application_config("telegraf")["install_method"][
            "value"
        ]

        if install_method == "deb":
            telegraf_conf = DEB_TELEGRAF_CONF
        else:
            telegraf_conf = SNAP_TELEGRAF_CONF
        cmd = "cat {}".format(telegraf_conf)
        response = model.run_on_unit(self.lead_unit_name, cmd)

        if response["Code"] != "0":
            self.fail(
                "test_02_telegraf_logfile: could not read file {}".format(telegraf_conf)
            )

        for line in response["Stdout"].splitlines():
            if line.strip() == 'logfile = "/var/log/telegraf/telegraf.log"':
                logging.info("test_02_telegraf_logfile: logfile config parameter found")

                return

        self.fail(
            "test_02_telegraf_logfile: logfile parameter not found in {}".format(
                telegraf_conf
            )
        )

    def test_03_system_service(self):
        """Check that the right service is running.

        e.g. either the deb's or the snap's.
        """
        install_method = model.get_application_config("telegraf")["install_method"][
            "value"
        ]
        services = {"deb": "telegraf", "snap": "snap.telegraf.telegraf"}

        for method in services.keys():
            service = services[method]
            cmd = "service {} status".format(service)
            response = model.run_on_unit(self.lead_unit_name, cmd)

            if install_method == method and response["Code"] != "0":
                self.fail(
                    "test_03_system_service: service {} should be running "
                    "on {} but is not. "
                    "install_method is {}.".format(
                        service, self.lead_unit_name, install_method
                    )
                )
            elif install_method != method and response["Code"] == "0":
                self.fail(
                    "test_03_system_service: service {}  is running "
                    "on {} but shouldn't. "
                    "install_method is {}.".format(
                        service, self.lead_unit_name, install_method
                    )
                )
